from typing import Dict
from typing import NoReturn
from typing import Type

from treat.mock.mock import Mock


class Mockery:
    def __init__(self) -> None:
        self._mocks: Dict[Type, Mock] = {}

    def mock(self, obj: Type) -> Mock:
        if obj in self._mocks:
            return self._mocks[obj]

        self._mocks[obj] = Mock(obj)

        return self._mocks[obj]

    def close(self) -> NoReturn:
        self._verify()

    def _verify(self) -> None:
        try:
            for mock in self._mocks.values():
                mock.verify()
        finally:
            self._reset()

    def _reset(self) -> None:
        for mock in self._mocks.values():
            mock.reset()

        self._mocks = {}



import re

from dataclasses import dataclass
from enum import Enum
from typing import Any
from typing import List
from typing import Optional
from typing import Tuple

from _pytest._code.code import ExceptionChainRepr
from _pytest.reports import TestReport


class TestResultStatus(Enum):
    FAILED = "failed"
    SKIPPED = "skipped"
    RUNS = "pending"
    PASSED = "passed"


@dataclass(frozen=True)
class TestActionResult:
    action: str
    status: TestResultStatus
    error: Optional[Exception] = None
    stdout: Optional[str] = None
    stderr: Optional[str] = None
    skip_reason: Optional[str] = None
    parameters: Optional[List[Tuple[str, Any]]] = None
    duration: Optional[float] = 0.0

    @classmethod
    def create_from_test_report(cls, report: TestReport) -> "TestActionResult":
        status = cls.make_status(report)
        action = report.when
        stdout = None
        if hasattr(report, "sections"):
            for name, content in report.sections:
                if "stdout" in name:
                    stdout = content

        skip_reason = None
        if report.outcome == TestResultStatus.SKIPPED.value:
            if hasattr(report, "wasxfail") and report.wasxfail:
                skip_reason = report.wasxfail
            elif isinstance(report.longrepr, ExceptionChainRepr):
                skip_reason = report.longrepr.chain[-1][2]
            else:
                skip_reason = report.longrepr[2][9:]

        return cls(
            action=action,
            status=status,
            error=report.exception if status == TestResultStatus.FAILED else None,
            stdout=stdout,
            skip_reason=skip_reason,
            parameters=list(report.parameters.items()),
            duration=report.duration,
        )

    @classmethod
    def make_status(cls, record: TestReport) -> TestResultStatus:
        if record.outcome == TestResultStatus.FAILED.value:
            return TestResultStatus.FAILED
        elif record.outcome == TestResultStatus.SKIPPED.value:
            return TestResultStatus.SKIPPED
        elif record.when in ["setup", "call"]:
            return TestResultStatus.RUNS

        return TestResultStatus.PASSED


class TestResult:
    FAIL = "failed"
    SKIPPED = "skipped"
    RUNS = "pending"
    PASS = "passed"

    def __init__(self, identifier: str, description: str) -> None:
        self._identifier = identifier
        self._description = description
        self._results: List[TestActionResult] = []

    @property
    def identifier(self) -> str:
        return self._identifier

    @property
    def base_identifier(self) -> str:
        return re.sub(r"\[(.+)]$", "", self._identifier)

    @property
    def description(self) -> str:
        return self._description

    @property
    def skip_reason(self) -> Optional[str]:
        return next(
            (r.skip_reason for r in self._results if r.skip_reason is not None), None
        )

    @property
    def status(self) -> TestResultStatus:
        if self.has_failed():
            return TestResultStatus.FAILED
        elif self.is_skipped():
            return TestResultStatus.SKIPPED
        elif self.is_running():
            return TestResultStatus.RUNS

        return TestResultStatus.PASSED

    @property
    def icon(self) -> str:
        return self._get_icon()

    @property
    def compact_icon(self) -> str:
        return self._get_icon(compact=True)

    @property
    def color(self) -> str:
        return self._get_color()

    @property
    def compact_color(self) -> str:
        return self._get_color(compact=True)

    @property
    def error(self) -> Optional[Exception]:
        return next((r.error for r in self._results if r.error is not None), None)

    @property
    def stdout(self) -> Optional[str]:
        if not self._results:
            return None

        return self._results[-1].stdout

    def has_failed(self) -> bool:
        return any(result.status is TestResultStatus.FAILED for result in self._results)

    def is_skipped(self) -> bool:
        return not self.has_failed() and any(
            result.status is TestResultStatus.SKIPPED for result in self._results
        )

    def is_running(self) -> bool:
        if not self._results:
            return True

        return self._results[-1].action != "teardown"

    @property
    def parameters(self) -> Optional[List[Tuple[str, Any]]]:
        if not self._results:
            return None

        return self._results[0].parameters

    @property
    def duration(self) -> float:
        return sum(result.duration for result in self._results)

    @classmethod
    def create_from_test_report(
        cls, report: TestReport, type: Optional[str] = None
    ) -> "TestResult":
        identifier = report.nodeid
        description = cls.make_description(report)

        result = TestResult(identifier, description)

        return result

    @classmethod
    def make_description(cls, record: TestReport) -> str:
        description = record.description
        if not description:
            description = record.nodeid.split("::")[-1]

            # Replace underscores with spaces
            description = description.replace("_", " ")

            # If it starts with `test_`, we remove it
            description = re.sub("^test(.*)", "\\1", description)

            # Remove spaces
            description = description.strip()

            # Drop parameters since they will be handled by the Style instance
            description = re.sub(r"\[(.+)]$", "", description)

        return description

    def add_result(self, result: TestActionResult) -> None:
        self._results.append(result)

    def _get_icon(self, compact: bool = False) -> str:
        status = self.status
        if status == TestResultStatus.FAILED:
            return "✕"
        elif status == TestResultStatus.SKIPPED:
            return "s"
        elif not self._results or self._results[-1].action in ["setup", "call"]:
            return "•" if not compact else "."

        return "✓" if not compact else "."

    def _get_color(self, compact: bool = False) -> str:
        status = self.status
        if status == TestResultStatus.FAILED:
            return "red"
        elif status == TestResultStatus.SKIPPED:
            return "yellow"
        elif not self._results or self._results[-1].action in ["setup", "call"]:
            return "blue" if not compact else "default;options=dark"

        return "green" if not compact else "default;options=dark"



import inspect

from enum import Enum
from typing import Any
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union

from treat.mock.exceptions import InvalidMockCallsCount


_UNSET = object()


class ExpectationKind(Enum):
    METHOD = "method"
    PROPERTY = "property"
    ATTRIBUTE = "attribute"


class Expectation:
    def __init__(
        self,
        name: str,
        kind: ExpectationKind,
    ) -> None:
        self._type = None
        self._name = name
        self._kind: ExpectationKind = kind
        self._times: Optional[int] = None
        self._arguments: Optional[Tuple] = None
        self._return_values: Union[object, List] = _UNSET
        self._calls = 0
        self._call_frame = None

    def times(self, times: int) -> "Expectation":
        self._times = times
        self._call_frame = inspect.stack()[1]

        return self

    def with_(self, *args, **kwargs) -> "Expectation":
        self._arguments = self._get_arguments(*args, **kwargs)

        return self

    def and_return(self, *return_values: Any) -> "Expectation":
        self._return_values = list(return_values)

        return self

    def verify(self) -> None:
        if self._times is not None and self._calls != self._times:
            raise InvalidMockCallsCount(
                self._name,
                self._times,
                self._calls,
                arguments=self._arguments,
                frame=self._call_frame,
            )

    def verify_call(self, *args, **kwargs) -> Any:
        self._increment_calls()

        if self._has_return_values():
            return self._consume_return_value()

    def match(self, *args, **kwargs) -> bool:
        arguments = self._get_arguments(*args, **kwargs)

        return self._arguments is None or arguments == self._arguments

    def _increment_calls(self, count: int = 1) -> "Expectation":
        self._calls += count

        return self

    def _has_return_values(self) -> bool:
        return self._return_values is not _UNSET

    def _consume_return_value(self) -> Any:
        value = self._return_values.pop(0)
        self._return_values.append(value)

        return value

    def _get_arguments(self, *args, **kwargs) -> Optional[Tuple]:
        if not args and not kwargs:
            return
        elif not kwargs:
            return args, None
        elif not args:
            return None, tuple(kwargs.items())
        else:
            return args, tuple(kwargs.items())



import types

from typing import TYPE_CHECKING
from typing import Any
from typing import Callable
from typing import List
from typing import Optional

from treat.mock.exceptions import MockError
from treat.mock.expectation import Expectation
from treat.mock.expectation import ExpectationKind


if TYPE_CHECKING:
    from treat.mock.mock import Mock


class ExpectationDirector:
    def __init__(self, name: str, mock: "Mock") -> None:
        if not hasattr(mock.mocked, name):
            raise MockError(f'{mock.mocked} does not have an attribute named "{name}".')

        self._name: str = name
        self._mock: "Mock" = mock
        self._original = object.__getattribute__(mock.mocked, name)
        self._expectations: List[Expectation] = []
        self._is_setup: bool = False
        self._local_override: bool = False

    def add_expectation(self, expectation: Expectation) -> None:
        self._expectations.append(expectation)

    def setup(self) -> None:
        if self._is_setup:
            return

        if issubclass(self._original.__class__, (types.FunctionType, types.MethodType)):
            self._set(
                types.MethodType(self.caller(ExpectationKind.METHOD), self._mock.mocked)
            )
        elif issubclass(self._original.__class__, property):
            self._set(self.caller(ExpectationKind.PROPERTY))
        else:
            self._set(self.caller(ExpectationKind.ATTRIBUTE))

    def teardown(self) -> None:
        self._unset()

    def caller(self, kind: ExpectationKind) -> Callable:
        if kind is ExpectationKind.METHOD:

            def caller(self_, *args, **kwargs):
                return self.call(*args, **kwargs)

        elif kind is ExpectationKind.PROPERTY:

            @property
            def caller(self_, *args, **kwargs):
                return self.call(self_, *args, **kwargs)

        else:

            @property
            def caller(self_, *args, **kwargs):
                return self.call(self_, *args, **kwargs)

        return caller

    def call(self, *args, **kwargs) -> Any:
        expectation = self.find_expectation(*args, **kwargs)

        if not expectation:
            return self._original(self._mock.mocked, *args, **kwargs)

        return expectation.verify_call(*args, **kwargs)

    def find_expectation(self, *args, **kwargs) -> Optional[Expectation]:
        for expectation in self._expectations:
            if expectation.match(*args, **kwargs):
                return expectation

    def verify(self) -> None:
        for expectation in self._expectations:
            expectation.verify()

    def _set(self, value: Any) -> None:
        local_override = False

        if (
            hasattr(self._mock.mocked, "__dict__")
            and type(self._mock.mocked.__dict__) is dict
        ):
            if self._name not in self._mock.mocked.__dict__:
                local_override = True

            self._mock.mocked.__dict__[self._name] = value
        else:
            setattr(self._mock.mocked, self._name, value)

        self._local_override = local_override

    def _unset(self) -> None:
        if (
            hasattr(self._mock.mocked, "__dict__")
            and self._name in self._mock.mocked.__dict__
            and type(self._mock.mocked.__dict__) is dict
        ):
            if self._local_override:
                del self._mock.mocked.__dict__[self._name]
            else:
                self._mock.mocked.__dict__[self._name] = self._original
        else:
            setattr(self._mock.mocked, self._name, self._original)



import inspect

from typing import Optional
from typing import Union

from cleo.io.io import IO
from cleo.io.outputs.output import Output
from cleo.ui.exception_trace import ExceptionTrace as _ExceptionTrace
from cleo.ui.exception_trace import Highlighter
from crashtest.frame import Frame
from crashtest.solution_providers.solution_provider_repository import (
    SolutionProviderRepository,
)

from treat.mock.exceptions import FrameAwareMockError
from treat.mock.exceptions import InvalidMockCallsCount

from .errors import UIError


class ExceptionTrace(_ExceptionTrace):
    def __init__(
        self,
        exception: Exception,
        solution_provider_repository: Optional[SolutionProviderRepository] = None,
        base_indent: int = 0,
    ) -> None:
        super().__init__(
            exception, solution_provider_repository=solution_provider_repository
        )

        self._base_indent = base_indent

    def _render_exception(self, io, exception):
        from crashtest.inspector import Inspector

        inspector = Inspector(exception)
        if not inspector.frames:
            return

        frames = inspector.frames
        # For mock errors we actually want to reference
        # the initial expection line instead of the internal
        # line where the expection is raised.
        # In order to do so, we get the initial frame of the expection
        # and add it to the stack trace.
        if (
            isinstance(self._exception, FrameAwareMockError)
            and self._exception.frame is not None
        ):
            frame = Frame(self._exception.frame)
            frames.append(frame)

        self._render_trace(io, frames)

        self._render_line(
            io, "<error>{}</error>".format(inspector.exception_name), True
        )
        io.write_line("")
        if isinstance(exception, UIError):
            exception_message = exception.pretty_error.replace("\n", "\n  ")
        elif isinstance(exception, AssertionError):
            exception_message = inspector.exception_message.replace("\n", "\n  ")
        else:
            exception_message = io.remove_format(inspector.exception_message).replace(
                "\n", "\n  "
            )

        self._render_line(io, "<b>{}</b>".format(exception_message))

        current_frame = inspector.frames[-1]
        self._render_snippet(
            io,
            current_frame,
            complete=isinstance(exception, (AssertionError, InvalidMockCallsCount)),
        )

        self._render_solution(io, inspector)

    def _render_snippet(
        self, io: Union[IO, Output], frame: Frame, complete: bool = False
    ):
        self._render_line(
            io,
            "at <fg=green>{}</>:<b>{}</b> in <fg=cyan>{}</>".format(
                self._get_relative_file_path(frame.filename),
                frame.lineno,
                frame.function,
            ),
            True,
        )

        if complete:
            start_line = frame.frame.f_code.co_firstlineno
            lines_before = frame.lineno - start_line + 1
            lines = inspect.getsourcelines(frame.frame.f_code)[0]
            lines_after = start_line + len(lines) - frame.lineno
        else:
            lines_before = lines_after = 4

        code_lines = Highlighter(supports_utf8=io.supports_utf8()).code_snippet(
            frame.file_content,
            frame.lineno,
            lines_before,
            lines_after,
        )

        for code_line in code_lines:
            self._render_line(io, code_line, indent=4)

    def _render_line(
        self, io: Output, line: str, new_line: bool = False, indent: int = 2
    ) -> None:
        indent += self._base_indent

        return super()._render_line(io, line, new_line=new_line, indent=indent)



from collections import defaultdict
from typing import Dict
from typing import Iterator
from typing import List
from typing import Optional
from typing import Tuple

from _pytest.reports import TestReport

from .test_result import TestActionResult
from .test_result import TestResult
from .test_result import TestResultStatus


class State:
    def __init__(self, group: Tuple[str]) -> None:
        self.suite_total_tests = None
        self._suite_tests: Dict[Tuple[str], Dict[str, TestResult]] = {}
        self._group = group
        self._group_tests: Dict[str, TestResult] = {}
        self._errors: Dict[Tuple[str], List[TestResult]] = {}
        self._errors_count = 0
        self._parameter_sets = defaultdict(dict)
        self._durations = defaultdict(lambda: 0.0)
        self.header_printed = False

    @property
    def group(self) -> Tuple[str]:
        return self._group

    @property
    def tests(self) -> Iterator[TestResult]:
        return self._group_tests.values()

    @property
    def group_title(self) -> str:
        if any(
            test.status == TestResultStatus.FAILED
            for test in self._group_tests.values()
        ):
            return "FAIL"

        if any(
            test.status != TestResultStatus.PASSED
            for test in self._group_tests.values()
        ):
            return "WARN"

        return "PASS"

    @property
    def group_title_color(self) -> str:
        if any(
            test.status == TestResultStatus.FAILED
            for test in self._group_tests.values()
        ):
            return "red"

        if any(
            test.status != TestResultStatus.PASSED
            for test in self._group_tests.values()
        ):
            return "yellow"

        return "green"

    @property
    def group_tests_count(self) -> int:
        return len(self._group_tests)

    @property
    def suite_tests_count(self) -> int:
        count = 0
        for group_tests in self._suite_tests.values():
            for _ in group_tests.values():
                count += 1

        return count

    @property
    def errors(self) -> Dict[Tuple[str], List[TestResult]]:
        return self._errors

    @property
    def errors_count(self) -> int:
        return self._errors_count

    def parameter_set(self, test) -> Optional[int]:
        if (
            test.base_identifier not in self._parameter_sets
            or test.identifier not in self._parameter_sets[test.base_identifier]
        ):
            return None

        return self._parameter_sets[test.base_identifier][test.identifier]

    def get_test_from_report(self, report: TestReport) -> TestResult:
        test = TestResult.create_from_test_report(report)

        if test.identifier in self._group_tests:
            return self._group_tests[test.identifier]

        return test

    def add_report(self, report: TestReport) -> None:
        test = self.get_test_from_report(report)

        self._group_tests[test.identifier] = test

        action_result = TestActionResult.create_from_test_report(report)
        test.add_result(action_result)

        if action_result.error:
            self._errors.setdefault(self._group, [])
            self._errors[self._group].append(test)
            self._errors_count += 1

        if self._group not in self._suite_tests:
            self._suite_tests[self._group] = {}

        self._suite_tests[self._group] = self._group_tests

        if action_result.parameters:
            if test.identifier not in self._parameter_sets[test.base_identifier]:
                self._parameter_sets[test.base_identifier][test.identifier] = (
                    len(self._parameter_sets[test.base_identifier]) + 1
                )

        self._durations[test.identifier] += action_result.duration

    def test_exists_in_group(self, test: TestResult) -> bool:
        return test.identifier in self._group_tests

    def group_has_changed(self, report: TestReport) -> bool:
        return tuple(report.nodeid.split("::")[:-1]) != self._group

    def move_to(self, report: TestReport) -> None:
        self._group = tuple(report.nodeid.split("::")[:-1])
        self._group_tests = {}
        self.header_printed = False

    def count_tests_in_suite_with_type(self, type: TestResultStatus) -> int:
        count = 0

        for group_tests in self._suite_tests.values():
            for test in group_tests.values():
                if test.status == type:
                    count += 1

        return count

    def duration(self, test: TestResult) -> float:
        return self._durations[test.identifier]
